/*
    ChibiOS/RT - Copyright (C) 2006,2007,2008,2009,2010,
                 2011,2012 Giovanni Di Sirio.

    This file is part of ChibiOS/RT.

    ChibiOS/RT is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    ChibiOS/RT is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

                                      ---

    A special exception to the GPL can be applied should you wish to distribute
    a combined work that includes ChibiOS/RT, without being obliged to provide
    the source code for any proprietary components. See the file exception.txt
    for full details of how and when the exception can be applied.
*/

#include "ch.h"
#include "hal.h"
#include "chprintf.h"
#include <ff.h>
#include <string.h>
#include "mad.h"

FATFS MMC_FS;
MMCDriver MMCD1;
static bool_t fs_ready = FALSE;
static SPIConfig hs_spicfg = { NULL, GPIOC, 4, 0 };
static SPIConfig ls_spicfg = { NULL, GPIOC, 4, SPI_CR1_BR_2 | SPI_CR1_BR_1 };
Thread *MadThd;

/* Card insertion verification.*/
static bool_t mmc_is_inserted(void)
{
  // not wired yet, assume card is always in
  //return palReadPad(IOPORT3, GPIOC_MMCCP);
  return 1;
}

static bool_t mmc_is_protected(void)
{
  // not wired yet
  //return !palReadPad(IOPORT3, GPIOC_MMCWP);
  return 0;
}

#include "mad.c"

FRESULT Mp3PlayAllFiles(char* path)
{
  FILINFO fno;
  DIR dir;
  char fpath[50];

  chprintf((BaseChannel*)&SD2, "Mp3PlayAllFiles: Playing all files in \"%s\"\r\n", path);
  FRESULT res = f_opendir(&dir, path);
  if(res == FR_OK)
  {
    res = f_readdir(&dir, &fno);
    while((res == FR_OK) && (fno.fname[0] != 0))
    {
      strcpy(fpath, path);
      strcat(fpath, fno.fname);
      if(fno.fattrib & AM_DIR)
      {
        chprintf((BaseChannel*)&SD2, "Mp3PlayAllFiles: DIR \"%s\"\r\n", fpath);
      }
      else
      {
        chprintf((BaseChannel*)&SD2, "Mp3PlayAllFiles: FILE \"%s\"\r\n", fpath);

        size_t unLen = strlen(fpath);
        if(unLen > 3)
        {
          if(stricmp(fpath + unLen - 4, ".mp3") == 0)
          {
	    decode(fpath);
          }
        }
      }
      res = f_readdir(&dir, &fno);
    }
  }
  else
  {
    chprintf((BaseChannel*)&SD2, "Mp3PlayAllFiles: f_opendir failed %d\r\n", res);
  }
  chprintf((BaseChannel*)&SD2, "Mp3PlayAllFiles: Finished playing all files\r\n", path);

  return res;
}

// MMC card insertion event
static void InsertHandler(eventid_t id)
{
  FRESULT err;
  (void) id;
  if(mmcConnect(&MMCD1))
  {
    chprintf((BaseChannel *) &SD2, "SD: Failed to connect to card\r\n");
    return;
  }
  else
  {
    chprintf((BaseChannel *) &SD2, "SD: Connected to card\r\n");
  }

  err = f_mount(0, &MMC_FS);
  if(err != FR_OK)
  {
    chprintf((BaseChannel *) &SD2, "SD: f_mount() failed %d\r\n", err);
    mmcDisconnect(&MMCD1);
    return;
  }
  else
  {
    chprintf((BaseChannel *) &SD2, "SD: File system mounted\r\n");
  }
  fs_ready = TRUE;

  // quick and dirty
  // directly decode all mp3's in the
  // cards root in a loop (assuming the card is never pulled)
  MadThd=chThdSelf();
  while(1)
  {
    Mp3PlayAllFiles("/");
  }
}

// MMC card removal event
static void RemoveHandler(eventid_t id)
{
  (void) id;
  fs_ready = FALSE;
}

// called on kernel panic
void port_halt(void)
{
  port_disable();
  while(TRUE)
  {
    palSetPad(GPIOD, GPIOD_LED4);       /* Green */
  }
}

void debug(void)
{
  chThdSleepMilliseconds(1000);

  chprintf((BaseChannel *)&SD2, "I2S:\r\n");
  chprintf((BaseChannel *)&SD2, " CR1 %x\r\n", SPI3->CR1);
  chprintf((BaseChannel *)&SD2, " CR2 %x\r\n", SPI3->CR2);
  chprintf((BaseChannel *)&SD2, " SR  %x\r\n", SPI3->SR);
  chprintf((BaseChannel *)&SD2, " CFG %x\r\n", SPI3->I2SCFGR);
  chprintf((BaseChannel *)&SD2, " PR  %x\r\n", SPI3->I2SPR);

  chprintf((BaseChannel *)&SD2, "DMA:\r\n");
  chprintf((BaseChannel *)&SD2, " CR  %x\r\n", SPID3.dmatx->stream->CR);
  chprintf((BaseChannel *)&SD2, " NDT %x\r\n", SPID3.dmatx->stream->NDTR);
  chprintf((BaseChannel *)&SD2, " PAR %x\r\n", SPID3.dmatx->stream->PAR);
  chprintf((BaseChannel *)&SD2, " M0A %x\r\n", SPID3.dmatx->stream->M0AR);
  chprintf((BaseChannel *)&SD2, " M1A %x\r\n", SPID3.dmatx->stream->M1AR);
  chprintf((BaseChannel *)&SD2, " FCR %x\r\n", SPID3.dmatx->stream->FCR);
}

/*
 * This is a periodic thread that does absolutely nothing except flashing
 * a LED.
 */
static WORKING_AREA(waThread1, 128);
__attribute__((noreturn)) static msg_t Thread1(void *arg) {

  (void)arg;
  chRegSetThreadName("blinker");
  while (TRUE) {
    palSetPad(GPIOD, GPIOD_LED3);       /* Orange.  */
    chThdSleepMilliseconds(500);
    palClearPad(GPIOD, GPIOD_LED3);     /* Orange.  */
    chThdSleepMilliseconds(500);
    debug();
  }
}

/*
 * Application entry point.
 */
int main(void) {
  static const evhandler_t evhndl[] = {InsertHandler, RemoveHandler};
  struct EventListener el0, el1;

  /*
   * System initializations.
   * - HAL initialization, this also initializes the configured device drivers
   *   and performs the board-specific initializations.
   * - Kernel initialization, the main() function becomes a thread and the
   *   RTOS is active.
   */
  halInit();
  chSysInit();

  /*
   * Activates the serial driver 2 using the driver default configuration.
   * PA2(TX) and PA3(RX) are routed to USART2.
   */
  sdStart(&SD2, NULL);
  palSetPadMode(GPIOA, 2, PAL_MODE_ALTERNATE(7));
  palSetPadMode(GPIOA, 3, PAL_MODE_ALTERNATE(7));

  // setup pads to SPI1 function (connect these pads to your SD card accordingly)
  palSetPadMode(GPIOC, 4, PAL_MODE_OUTPUT_PUSHPULL | PAL_STM32_OSPEED_HIGHEST); // NSS
  palSetPadMode(GPIOA, 5, PAL_MODE_ALTERNATE(5) | PAL_STM32_OSPEED_HIGHEST); // SCK
  palSetPadMode(GPIOA, 6, PAL_MODE_ALTERNATE(5)); // MISO
  palSetPadMode(GPIOA, 7, PAL_MODE_ALTERNATE(5) | PAL_STM32_OSPEED_HIGHEST); // MOSI
  palSetPad(GPIOC, 4); // set NSS high

  // initialize MMC driver
  mmcObjectInit(&MMCD1, &SPID1, &ls_spicfg, &hs_spicfg, mmc_is_protected, mmc_is_inserted);
  mmcStart(&MMCD1, NULL);

  /*
   * Creates the example thread.
   */
  chThdCreateStatic(waThread1, sizeof(waThread1), NORMALPRIO, Thread1, NULL);

  EVAL_AUDIO_Init(0,70,44100);

  chprintf((BaseChannel *)&SD2, "STM32 Audio MP3 / ChibiOS / Libmad : %s %s %s\r\n", __FILE__, __DATE__, __TIME__);
  chprintf((BaseChannel *)&SD2, "%s\r\n", mad_version);
  chprintf((BaseChannel *)&SD2, "%s\r\n", mad_copyright);
  chprintf((BaseChannel *)&SD2, "%s\r\n", mad_author);
  chprintf((BaseChannel *)&SD2, "%s\r\n", mad_build);

  // GO

  chEvtRegister(&MMCD1.inserted_event, &el0, 0);
  chEvtRegister(&MMCD1.removed_event, &el1, 1);
  while(TRUE)
  {
    chEvtDispatch(evhndl, chEvtWaitOne(ALL_EVENTS));
  }
}
